'QuickBasic should be started with the command line
'qb /ah /lallmodex
'In order to use the Mode-X libraries.

'$DYNAMIC
'do not remove the DYNAMIC metacommand, it is not really a remark. It is
'actually a compiler directive that improves memory management.

DEFINT A-Z
'Because all of these commands use integer arguments, it may be a good idea
'to make integer variables the default. Then it will assume that any variable
'that does not have a symbol after it will be treated like it has a % after
'it. Remember that if you do this, you must put a ! at the end of float
'variables

'---------------------------------------------------------------------------
'The following are the declarations for the assembly-code routines. You
'only need the ones you are actually going to use.

DECLARE SUB setmodex ()
'Puts the screen into Mode-X

DECLARE SUB restoremode ()
'Returns the screen into whatever mode it was before you called setmodex.
'Note that if you run setmodex twice, and then run restoremode, you will
'still be in mode-x. restoremode only remebers the most recent screen mode.

DECLARE SUB setvispage (BYVAL pagenumber%)
DECLARE SUB clearpage (BYVAL pagenumber%)
DECLARE SUB copypage (BYVAL sourcepage%, BYVAL destpage%)
'Mode-X is divided into 4 screen pages, just like EGA mode-7. Only one page
'is visable at a time. The pages are numbered 0,1,2,3

DECLARE SUB setpal (pal())
'This routine loads a custom 256-color master-palette from a 768 element
'array.
'  DIM masterpal(768)
'  DEF SEG = VARSEG(masterpal(0)): BLOAD "sample.mas", VARPTR(masterpal(0))
'  setpal masterpal()

DECLARE SUB setfont (font%())
'Mode-X cannot use normal text. Instead you have load a font.
'setfont loads a font from a 1024 element integer array
'  DIM font(1024)
'  DEF SEG = VARSEG(font(0)): BLOAD "sample.fnt", VARPTR(font(0))
'  setfont font()

DECLARE SUB textcolor (BYVAL fg%, BYVAL bg%)
DECLARE SUB printstr (text$, BYVAL x%, BYVAL y%, BYVAL page%)
'textcolor sets the color will be drawn in. fg% is the forground color
'and bg% is the background color. A value of 0 for either is transparent.
'printstr is the mode-x replacement for PRINT. Unlike PRINT, printstr
'must know what X,Y, and page you want to put it on. The x,y coordinates
'are in pixels, not in characters.

DECLARE SUB putpixel (BYVAL x%, BYVAL y%, BYVAL col%, BYVAL page%)
'the putpixel command draws a single pixel of a given color at x,y on page

DECLARE FUNCTION readpixel% (BYVAL x%, BYVAL y%, BYVAL page%)
'this function returns the color of the pixel at x,y on page

DECLARE SUB rectangle (BYVAL x%, BYVAL y%, BYVAL w%, BYVAL h%, BYVAL col%, BYVAL page%)
DECLARE SUB fuzzyrect (BYVAL x%, BYVAL y%, BYVAL w%, BYVAL h%, BYVAL col%, BYVAL page%)
'draw a rectangle. x,y are the top left corner. w,h are the width and height.
'col is the color, and page is the page to draw on.
'fuzzyrect is the same as rectangle, but only draws every-other pixel

DECLARE SUB drawline (BYVAL x1%, BYVAL y1%, BYVAL x2%, BYVAL y2%, BYVAL col%, BYVAL page%)
'drawline draws a line of color col% from x1,y1 to x2,y2 on page%

DECLARE SUB paintat (BYVAL x%, BYVAL y%, BYVAL col%, BYVAL page%, buffer%(), BYVAL bufsize%)
'paintat fills an area starting at x,y with a color.
'buffer%() must be at least as big as the area you are painting.
'bufsize% is the size of the buffer in bytes (DIMs * 2)

DECLARE FUNCTION bmpinfo (f$, dat())
'this function attempts to read some info about a BMP file.
'f$ must be zero-terminated. dat() is an X byte that is filled with
'information from the BMP's header.
'(0) bit-depth
'(1) x-width in pixels
'(2) y-height in pixels
'(3+) data directly from the header, for as many bytes long as your buffer is
'When bmpinfo is run, it tries to open f$. If it fails, or if the
'file does not have a BMP header, bmpinfo will return false.

DECLARE SUB loadbmp (f$, BYVAL x, BYVAL y, buf(), BYVAL page)
'loads a 4-bit BMP file named by the zero-terminated string f$
'and writes it at position X,Y on the specified screen page.
'the buffer needs to be at least width*height/4 bytes... I think :)
'this data is written using the first 16 colors of the master palette, so
'it is suitable for loading with getsprite

DECLARE SUB getbmppal (f$, masterpal(), pal(), BYVAL offset)
'loads the 16-color palette out of the BMP file named in the zero-terminated
'string f$, and saves it into the palette array pal() at the given offset
'it is neccisary to give an array that contains the currently loaded
'master palette so that the colors can be resolved.

DECLARE SUB drawsprite (pic(), BYVAL picoffset, pal(), BYVAL paloffset, BYVAL x, BYVAL y, BYVAL page)
DECLARE SUB wardsprite (pic(), BYVAL picoffset, pal(), BYVAL paloffset, BYVAL x, BYVAL y, BYVAL page)
DECLARE SUB bigsprite (pic(), pal(), BYVAL paloffset, BYVAL x, BYVAL y, BYVAL page)
DECLARE SUB hugesprite (pic(), pal(), BYVAL paloffset, BYVAL x, BYVAL y, BYVAL page)
'drawsprite and it's variations all draw a 16-color picture from a buffer
'onto a screen page using the specified palette. the picoffset tells what
'element of the pic() buffer the sprite starts at, and the paloffset tells
'what element of the pal() buffer the palette starts at
'wardsprite mirrors the sprite over the Y axis
'bigsprite and hugesprite draw zoomed versions

DECLARE SUB getsprite (pic(), BYVAL picoff, BYVAL x, BYVAL y, BYVAL wide, BYVAL high, BYVAL page)
'reads a rectangular region of a screen page into a buffer as if it is a
'sprite. This region should be drawn only in the first 16 colors of the
'master palette (getsprite is most useful for reading data that has
'been imported with loadbmp so that you can drawsprite it)

DECLARE SUB stosprite (pic(), BYVAL picoffset, BYVAL x, BYVAL y, BYVAL page)
'writes a sprite from a buffer to a screen page as raw data for storage

DECLARE SUB loadsprite (pic(), BYVAL picoffset, BYVAL x, BYVAL y, BYVAL wide, BYVAL high, BYVAL page)
'reads raw data into a buffer from a screen page. This is used to reload
'data that has been written with stosprite. It is neccisary to specify the
'width and height of the sprite, because that information is not included
'in the raw data

DECLARE SUB bitmap2page (pal(), bmp$, BYVAL page)
'loads a 320x200 pixel 24-bit BMP file to a screen page, and fits it to
'your current 8-bit palette. pal() is your master palette. bmp$ is the
'zero-terminated name of the bitmap, and page is the distination page.

DECLARE SUB setupmusic (mbuf())
'setupmusic prepares a buffer that your music will use to play.
'run it once at the very beginning of any program where you want to
'play BAM music. mbuf() is an array that needs to be big enough to
'hold the largest BAM file you will be playing (remeber that an array
'of integers holds 2 bytes per element, so if you DIM mbuf(16000) you
'will be able to play any BAM up to 32k. that is a good number, since it
'is the largest file size that either MIDI2BAM.EXE or NOTATE.EXE can
'produce

DECLARE SUB closemusic ()
'run this once at the end of your program to de-allocate the buffer you
'set up with setupmusic

DECLARE SUB resetfm ()
'reset the FM chip in your sound card. Call this once right after
'setupmusic.

DECLARE SUB loadsong (f$)
'loads a BAM file into memory and starts it playing. f$ is the file name,
'and should be zero-terminated. Example:
'loadsong "MYSONG.BAM"+CHR$(0)

DECLARE SUB stopsong ()
'stop the currently playing song. use it at any time in your program

DECLARE SUB resumesong ()
'restarts the music after you stopped with stopsong

DECLARE FUNCTION getfmvol ()
'returns the current FM mixer volume. It will be a number between 0 and 15
'on most sound cards, or between 0 and 7 for some older cards

DECLARE SUB setfmvol (BYVAL vol)
'sets the FM mixer volume to the target volume. vol is a number between
'0 and 15

DECLARE SUB fademusic (BYVAL vol)
'fades the FM volume towards a target volume.


'The following routines are also available, but I havent had time to
'write explantions of them yet. I will as time allows.

DECLARE SUB fadeto (palbuff(), BYVAL red, BYVAL green, BYVAL blue)
DECLARE SUB fadetopal (pal(), palbuff())
DECLARE SUB setdiskpages (buf(), BYVAL h, BYVAL l)
DECLARE SUB loadpage (fil$, BYVAL i, BYVAL p)
DECLARE SUB storepage (fil$, BYVAL i, BYVAL p)

DECLARE SUB setbit (b(), BYVAL w, BYVAL b, BYVAL v)
DECLARE FUNCTION readbit (b(), BYVAL w, BYVAL b)

'--registers type for calling interrupts
TYPE Regtype
 ax AS INTEGER
 bx AS INTEGER
 cx AS INTEGER
 dx AS INTEGER
 bp AS INTEGER
 si AS INTEGER
 di AS INTEGER
 flags AS INTEGER
 ds AS INTEGER
 es AS INTEGER
END TYPE
DIM SHARED regs AS Regtype

'---------------------------------------------------------------------------
'put your actual program _after_ all of the DECLARE statements

'--arrays
DIM masterpal(768)   'holds the master palette
DIM font(1024)       'holds the font
DIM bmpheader(2)     'for use with bmpinfo
DIM fourbitpal(32)   'holds 16-color palettes. this buffer has room for
                     'two of them
DIM drawbuffer(2502) 'this buffer is big enough to hold any 4-bit image
                     'up to 100*100 pixels in size. to calculate the size of
                     'your drawing buffer multiply width times height and
                     'divide by 4 and add 2 (the extra 2 is to store the
                     'height and width)

'--change the screen mode to Mode-X
setmodex
setvispage 0

'--load the master palette
DEF SEG = VARSEG(masterpal(0)): BLOAD "ohrrpgce.mas", VARPTR(masterpal(0))
setpal masterpal()

'--load rom font
regs.ax = &H1130
regs.bx = &H300
CALL interruptx(&H10, regs, regs)
off9 = regs.bx: seg9 = regs.es
DEF SEG = regs.es
FOR i = 1 TO 255
  FOR j = 0 TO 7
    b = PEEK(regs.bp + (8 * i) + j)
    FOR k = 0 TO 7
      setbit font(), i * 4, (7 - k) * 8 + j, (b AND 2 ^ k)
    NEXT k
  NEXT j
NEXT i

setfont font()

'--white text, transparent background
textcolor 15, 0
printstr "Welcome to BriTech Mode-X", 0, 0, 0
w$ = INPUT$(1)
clearpage 0

'---------------------------------------------------------------------------

printstr "draw out a pixel of each color", 0, 0, 0
w$ = INPUT$(1)
clearpage 0

FOR i = 0 TO 255
 putpixel (i MOD 16), INT(i / 16), i, 0
NEXT i
printstr "this is the palette", 16, 16, 0

w$ = INPUT$(1)
clearpage 0

'---------------------------------------------------------------------------

printstr "display a sample bitmap", 0, 0, 0
w$ = INPUT$(1)
clearpage 0

bmp$ = "importx\enemy\armarm.BMP"

'--get info on sample bitmap
IF bmpinfo(bmp$ + CHR$(0), bmpheader()) THEN
  printstr "header info for " + bmp$, 0, 0, 0
  printstr "bit depth" + STR$(bmpheader(0)), 0, 8, 0
  printstr "width" + STR$(bmpheader(1)), 0, 16, 0
  printstr "height" + STR$(bmpheader(2)), 0, 24, 0
ELSE
  printstr "unable to open or not a bitmap", 0, 0, 0
  '--if bmpinfo fails, loadbmp will certainly fail too
END IF

w$ = INPUT$(1)
clearpage 0

'--load the bitmap and store it temporaraly on screen page 3
loadbmp bmp$ + CHR$(0), 0, 0, drawbuffer(), 3

'--now we will load the picture into the buffer so that drawsprite can
'display it
getsprite drawbuffer(), 0, 0, 0, bmpheader(1), bmpheader(2), 3

'--load the bitmap's palette into the first 16 elements of fourbitpal()
getbmppal bmp$ + CHR$(0), masterpal(), fourbitpal(), 0

drawsprite drawbuffer(), 0, fourbitpal(), 0, 10, 10, 0

w$ = INPUT$(1)
clearpage 0

'---------------------------------------------------------------------------

'-- lets do it again with a different BMP file

printstr "display another sample bitmap", 0, 0, 0
w$ = INPUT$(1)
clearpage 0

bmp$ = "importx\enemy\red_plip.BMP"

IF NOT bmpinfo(bmp$ + CHR$(0), bmpheader()) THEN
  printstr "unable to open or not a bitmap", 0, 0, 0
ELSE
 clearpage 3
 loadbmp bmp$ + CHR$(0), 0, 0, drawbuffer(), 3
 getsprite drawbuffer(), 0, 0, 0, bmpheader(1), bmpheader(2), 3
 getbmppal bmp$ + CHR$(0), masterpal(), fourbitpal(), 16
 drawsprite drawbuffer(), 0, fourbitpal(), 16, 10, 10, 0
END IF

w$ = INPUT$(1)
clearpage 0

'---------------------------------------------------------------------------

'--now lets try a full-screen bitmap

printstr "load and color-fit a full-screen", 0, 0, 0
printstr "320*200 pixel 24-bit bitmap.", 0, 8, 0
w$ = INPUT$(1)
clearpage 0

bmp$ = "import\sandsea.BMP"

IF NOT bmpinfo(bmp$ + CHR$(0), bmpheader()) THEN
 printstr "unable to open or not a bitmap", 0, 0, 0
ELSE
 IF bmpheader(0) = 24 THEN
  bitmap2page masterpal(), bmp$ + CHR$(0), 0
 ELSE
  printstr "bitmap is not 24-bit color", 0, 0, 0
 END IF
END IF

w$ = INPUT$(1)

'---------------------------------------------------------------------------

'--as you may have noticed, loading that full screen bitmap was kinda slow.
'--after you have it converted into your mode-x palette, you can save it as
'--a mode-x screen.



'---------------------------------------------------------------------------

'--revert the screen to the previous screen mode
restoremode
SYSTEM

